/*
 *  HID driver for N-Trig touchscreens
 *
 *  Copyright (c) 2008-2010 Rafi Rubin
 *  Copyright (c) 2009-2010 Stephane Chatty
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 */

#include <linux/device.h>
#include <linux/hid.h>
#include <linux/usb.h>
#include "usbhid/usbhid.h"
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/input/mt.h>

#include "hid-ids.h"

#define NTRIG_DUPLICATE_USAGES	0x001
/**
 * list_rotate_right - rotate the list to the right
 * @head: the head of the list
 */
static inline void list_rotate_right(struct list_head *head)
{
        struct list_head *last;

        if (!list_empty(head)) {
                last = head->prev;
                list_move(last, head);
        }
}

static const char *ntrig_modes[4] = { "pen", "touch", "auto", "dual" };

static unsigned int min_width;
module_param(min_width, uint, 0644);
MODULE_PARM_DESC(min_width, "Minimum touch contact width to accept.");

static unsigned int min_height;
module_param(min_height, uint, 0644);
MODULE_PARM_DESC(min_height, "Minimum touch contact height to accept.");

static unsigned int activate_slack = 1;
module_param(activate_slack, uint, 0644);
MODULE_PARM_DESC(activate_slack, "Number of touch frames to ignore at "
		 "the start of touch input.");

static unsigned int deactivate_slack = 4;
module_param(deactivate_slack, uint, 0644);
MODULE_PARM_DESC(deactivate_slack, "Number of empty frames to ignore before "
		 "deactivating touch.");

static unsigned int activation_width = 64;
module_param(activation_width, uint, 0644);
MODULE_PARM_DESC(activation_width, "Width threshold to immediately start "
		 "processing touch events.");

static unsigned int activation_height = 32;
module_param(activation_height, uint, 0644);
MODULE_PARM_DESC(activation_height, "Height threshold to immediately start "
		 "processing touch events.");

struct ntrig_slot {
	int id;
	struct list_head list;
};

struct ntrig_contact {
	__u16 x, y, w, h;
	__u16 id;
	__s16 est_x_min, est_x_max, est_y_min, est_y_max;
	__s16 dx, dy;

	/* An age factor for counting frames since a track was last seen.
	 * This enables drop compensation and delayed termination. */
	__u8 inactive;

	struct ntrig_slot *slot;

	struct list_head list;

	/* List of tracks sorted by first seen order */
	struct list_head active_tracks;
};

struct ntrig_frame {

	/* Items that represent physical contacts which have been mapped
	 * to contacts from previous frames. */
	struct list_head tracked;

	/* Contacts that have yet to be matched and might be ghosts. */
	struct list_head pending;
	struct list_head list;
};

struct ntrig_data {
	/* Incoming raw values for a single contact */
	__u16 x, y, w, h;
	__u16 id;
	//int slots;

	bool tipswitch;
	bool confidence;
	bool first_contact_touch;

	bool reading_mt;

	__u8 maxcontacts;
	__u8 mt_footer[4];
	__u8 mt_foot_count;

	/* The current activation state. */
	__s8 act_state;

	/* Empty frames to ignore before recognizing the end of activity */
	__s8 deactivate_slack;

	/* Frames to ignore before acknowledging the start of activity */
	__s8 activate_slack;

	/* Minimum size contact to accept */
	__u16 min_width;
	__u16 min_height;

	/* Threshold to override activation slack */
	__u16 activation_width;
	__u16 activation_height;

	__u16 sensor_logical_width;
	__u16 sensor_logical_height;
	__u16 sensor_physical_width;
	__u16 sensor_physical_height;

	__u16 r_y;
	__u16 r_x;

	/* Circular list of frames used to maintain state of contacts */
	struct list_head frames;

	/* Contacts representing the last input from lost tracks and
	 * old contacts to be recycled */
	struct list_head old_contacts;

	struct list_head available_slots;
	struct list_head active_tracks;

	struct ntrig_frame *first_frame;
	struct ntrig_contact *first_contact;
	struct ntrig_slot *first_slot;
};


/*
 * This function converts the 4 byte raw firmware code into
 * a string containing 5 comma separated numbers.
 */
static int ntrig_version_string(unsigned char *raw, char *buf)
{
	__u8 a =  (raw[1] & 0x1e) >> 1;
	__u8 b =  (raw[0] & 0x3c) >> 2;
	__u8 c = ((raw[0] & 0x03) << 3) | ((raw[3] & 0xe0) >> 5);
	__u8 d = ((raw[3] & 0x07) << 3) | ((raw[2] & 0xe0) >> 5);
	__u8 e =   raw[2] & 0x1f;

	/*
	 * As yet unmapped bits:
	 * 0b11000000 0b11100001 0b00000000 0b00011000
	 */

	return sprintf(buf, "%u.%u.%u.%u.%u", a, b, c, d, e);
}

static inline int ntrig_get_mode(struct hid_device *hdev)
{
	struct hid_report *report = hdev->report_enum[HID_FEATURE_REPORT].
				    report_id_hash[0x0d];

	if (!report)
		return -EINVAL;

	usbhid_submit_report(hdev, report, USB_DIR_IN);
	usbhid_wait_io(hdev);
	return (int)report->field[0]->value[0];
}

static inline void ntrig_set_mode(struct hid_device *hdev, const int mode)
{
	struct hid_report *report;
	__u8 mode_commands[4] = { 0xe, 0xf, 0x1b, 0x10 };

	if (mode < 0 || mode > 3)
		return;

	report = hdev->report_enum[HID_FEATURE_REPORT].
		 report_id_hash[mode_commands[mode]];

	if (!report)
		return;

	usbhid_submit_report(hdev, report, USB_DIR_IN);
}

static void ntrig_report_version(struct hid_device *hdev)
{
	int ret;
	char buf[20];
	struct usb_device *usb_dev = hid_to_usb_dev(hdev);
	unsigned char *data = kmalloc(8, GFP_KERNEL);

	if (!data)
		goto err_free;

	ret = usb_control_msg(usb_dev, usb_rcvctrlpipe(usb_dev, 0),
			      USB_REQ_CLEAR_FEATURE,
			      USB_TYPE_CLASS | USB_RECIP_INTERFACE |
			      USB_DIR_IN,
			      0x30c, 1, data, 8,
			      USB_CTRL_SET_TIMEOUT);

	if (ret == 8) {
		ret = ntrig_version_string(&data[2], buf);

		hid_info(hdev, "Firmware version: %s (%02x%02x %02x%02x)\n",
			 buf, data[2], data[3], data[4], data[5]);
	}

err_free:
	kfree(data);
}

static ssize_t show_min_width(struct device *dev,
			      struct device_attribute *attr,
			      char *buf)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	return sprintf(buf, "%d\n", nd->min_width *
				    nd->sensor_physical_width /
				    nd->sensor_logical_width);
}

static ssize_t set_min_width(struct device *dev,
			     struct device_attribute *attr,
			     const char *buf, size_t count)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	unsigned long val;

	if (strict_strtoul(buf, 0, &val))
		return -EINVAL;

	if (val > nd->sensor_physical_width)
		return -EINVAL;

	nd->min_width = val * nd->sensor_logical_width /
			      nd->sensor_physical_width;

	return count;
}

static DEVICE_ATTR(min_width, S_IWUSR | S_IRUGO, show_min_width, set_min_width);

static ssize_t show_min_height(struct device *dev,
			       struct device_attribute *attr,
			       char *buf)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	return sprintf(buf, "%d\n", nd->min_height *
				    nd->sensor_physical_height /
				    nd->sensor_logical_height);
}

static ssize_t set_min_height(struct device *dev,
			      struct device_attribute *attr,
			      const char *buf, size_t count)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	unsigned long val;

	if (strict_strtoul(buf, 0, &val))
		return -EINVAL;

	if (val > nd->sensor_physical_height)
		return -EINVAL;

	nd->min_height = val * nd->sensor_logical_height /
			       nd->sensor_physical_height;

	return count;
}

static DEVICE_ATTR(min_height, S_IWUSR | S_IRUGO, show_min_height,
		   set_min_height);

static ssize_t show_activate_slack(struct device *dev,
				   struct device_attribute *attr,
				   char *buf)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	return sprintf(buf, "%d\n", nd->activate_slack);
}

static ssize_t set_activate_slack(struct device *dev,
				  struct device_attribute *attr,
				  const char *buf, size_t count)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	unsigned long val;

	if (strict_strtoul(buf, 0, &val))
		return -EINVAL;

	if (val > 0x7f)
		return -EINVAL;

	nd->activate_slack = val;

	return count;
}

static DEVICE_ATTR(activate_slack, S_IWUSR | S_IRUGO, show_activate_slack,
		   set_activate_slack);

static ssize_t show_activation_width(struct device *dev,
				     struct device_attribute *attr,
				     char *buf)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	return sprintf(buf, "%d\n", nd->activation_width *
				    nd->sensor_physical_width /
				    nd->sensor_logical_width);
}

static ssize_t set_activation_width(struct device *dev,
				    struct device_attribute *attr,
				    const char *buf, size_t count)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	unsigned long val;

	if (strict_strtoul(buf, 0, &val))
		return -EINVAL;

	if (val > nd->sensor_physical_width)
		return -EINVAL;

	nd->activation_width = val * nd->sensor_logical_width /
				     nd->sensor_physical_width;

	return count;
}

static DEVICE_ATTR(activation_width, S_IWUSR | S_IRUGO, show_activation_width,
		   set_activation_width);

static ssize_t show_activation_height(struct device *dev,
				      struct device_attribute *attr,
				      char *buf)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	return sprintf(buf, "%d\n", nd->activation_height *
				    nd->sensor_physical_height /
				    nd->sensor_logical_height);
}

static ssize_t set_activation_height(struct device *dev,
				     struct device_attribute *attr,
				     const char *buf, size_t count)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	unsigned long val;

	if (strict_strtoul(buf, 0, &val))
		return -EINVAL;

	if (val > nd->sensor_physical_height)
		return -EINVAL;

	nd->activation_height = val * nd->sensor_logical_height /
				      nd->sensor_physical_height;

	return count;
}

static DEVICE_ATTR(activation_height, S_IWUSR | S_IRUGO,
		   show_activation_height, set_activation_height);

static ssize_t show_deactivate_slack(struct device *dev,
				     struct device_attribute *attr,
				     char *buf)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	return sprintf(buf, "%d\n", -nd->deactivate_slack);
}

static ssize_t set_deactivate_slack(struct device *dev,
				    struct device_attribute *attr,
				    const char *buf, size_t count)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	unsigned long val;

	if (strict_strtoul(buf, 0, &val))
		return -EINVAL;

	/*
	 * No more than 8 terminal frames have been observed so far
	 * and higher slack is highly likely to leave the single
	 * touch emulation stuck down.
	 */
	if (val > 7)
		return -EINVAL;

	nd->deactivate_slack = -val;

	return count;
}

static DEVICE_ATTR(deactivate_slack, S_IWUSR | S_IRUGO, show_deactivate_slack,
		   set_deactivate_slack);


static ssize_t show_mode(struct device *dev,
			struct device_attribute *attr,
			char *buf)
{
	int mode = ntrig_get_mode(container_of(dev, struct hid_device, dev));
	int i, ret;
	char *s = buf;

	if (mode < 0)
		return mode;


	for (i = 0; i < 4; i++)
		s += sprintf(s, "%s%s%s ", (i == mode) ? "[" : "",
			     ntrig_modes[i], (i == mode) ? "]" : "");

	if (mode >= 4)
		s += sprintf(s, "[%d]\n", mode);
	else
		*(s - 1) = '\n';

	ret = s - buf;

	return ret;
}

static ssize_t store_mode(struct device *dev,
			       struct device_attribute *attr,
			       const char *buf, size_t count)
{
	int len;
	int i;
	int mode = -EINVAL;
	char *p;

	p = memchr(buf, '\n', count);
	len = p ? p - buf : count;

	if (len == 1 && buf[0] >= '0' && buf[0] <= '4') {
		mode = buf[0] - '0';
	} else
		for (i = 0; i < 4; i++)
			if (len == strlen(ntrig_modes[i]) &&
			    !strncmp(buf, ntrig_modes[i], len)) {
				mode = i;
				break;
			}

	if (mode < 0)
		return mode;

	ntrig_set_mode(container_of(dev, struct hid_device, dev), mode);

	return count;
}
static DEVICE_ATTR(mode, S_IWUSR | S_IRUGO, show_mode, store_mode);

static ssize_t ntrig_calibrate(struct device *dev,
			       struct device_attribute *attr,
			       const char *buf, size_t count)
{
	struct hid_device *hdev = container_of(dev, struct hid_device, dev);
	struct usb_device *usb_dev = hid_to_usb_dev(hdev);
	struct usbhid_device *usbhid = hdev->driver_data;
	int ret;
	unsigned long t;
	unsigned char *data;

	if (strict_strtoul(buf, 0, &t))
		return -EINVAL;

	data = kmalloc(4, GFP_KERNEL);
	if (!data)
		return -ENOMEM;

	spin_lock(&usbhid->lock);
	set_bit(HID_DISCONNECTED, &usbhid->iofl);
	spin_unlock(&usbhid->lock);

	ret = usb_control_msg(usb_dev, usb_rcvctrlpipe(usb_dev, 0),
			      USB_REQ_CLEAR_FEATURE, USB_TYPE_CLASS
			      | USB_RECIP_INTERFACE | USB_DIR_IN,
			      0x30b, 1, data, 4, USB_CTRL_GET_TIMEOUT);
	if (ret < 0)
		goto fail;

	msleep(t);

	ret = usb_control_msg(usb_dev, usb_rcvctrlpipe(usb_dev, 0),
			      USB_REQ_CLEAR_FEATURE, USB_TYPE_CLASS
			      | USB_RECIP_INTERFACE | USB_DIR_IN,
			      0x311, 1, data, 4, USB_CTRL_GET_TIMEOUT);

fail:
	kfree(data);
	spin_lock(&usbhid->lock);
	clear_bit(HID_DISCONNECTED, &usbhid->iofl);
	spin_unlock(&usbhid->lock);
	schedule_work(&usbhid->reset_work);

	return (ret < 0) ? ret : count;
}
static DEVICE_ATTR(calibrate, S_IWUSR, NULL, ntrig_calibrate);

static struct attribute *sysfs_attrs[] = {
	&dev_attr_min_height.attr,
	&dev_attr_min_width.attr,
	&dev_attr_activate_slack.attr,
	&dev_attr_activation_width.attr,
	&dev_attr_activation_height.attr,
	&dev_attr_deactivate_slack.attr,
	&dev_attr_mode.attr,
	&dev_attr_calibrate.attr,
	NULL
};

static struct attribute_group ntrig_attribute_group = {
	.attrs = sysfs_attrs
};

/*
 * this driver is aimed at two firmware versions in circulation:
 *  - dual pen/finger single touch
 *  - finger multitouch, pen not working
 */

static int ntrig_input_mapping(struct hid_device *hdev, struct hid_input *hi,
			       struct hid_field *field, struct hid_usage *usage,
			       unsigned long **bit, int *max)
{
	struct ntrig_data *nd = hid_get_drvdata(hdev);

	/* logical span */
	int dl = field->logical_maximum - field->logical_minimum;

	/* physical spam (10 points/mm) */
	int dp = field->physical_maximum - field->physical_minimum;

	/* No special mappings needed for the pen and single touch */
	if (field->physical)
		return 0;

	switch (usage->hid & HID_USAGE_PAGE) {
	case HID_UP_GENDESK:
		switch (usage->hid) {
		case HID_GD_X:
			hid_map_usage(hi, usage, bit, max,
					EV_ABS, ABS_MT_POSITION_X);
			input_set_abs_params(hi->input, ABS_MT_POSITION_X,
					     field->logical_minimum,
					     field->logical_maximum,
					     /* 2mm fuzz */ 20 * dl / dp, 0);
			input_set_abs_params(hi->input, ABS_X,
					     field->logical_minimum,
					     field->logical_maximum,
					     /* 2mm fuzz */ 20 * dl / dp, 0);

			if (!nd->sensor_logical_width) {
				nd->sensor_logical_width = dl;
				nd->sensor_physical_width = dp;
				nd->activation_width = activation_width *
						       dl / dp;
				nd->min_width = min_width * dl / dp;
			}
			return 1;
		case HID_GD_Y:
			hid_map_usage(hi, usage, bit, max,
				      EV_ABS, ABS_MT_POSITION_Y);
			input_set_abs_params(hi->input, ABS_MT_POSITION_Y,
					     field->logical_minimum,
					     field->logical_maximum,
					     /* 2mm fuzz */ 20 * dl / dp, 0);
			input_set_abs_params(hi->input, ABS_Y,
					     field->logical_minimum,
					     field->logical_maximum,
					     /* 2mm fuzz */ 20 * dl / dp, 0);

			if (!nd->sensor_logical_height) {
				nd->sensor_logical_height = dl;
				nd->sensor_physical_height = dp;
				nd->activation_height = activation_height *
							dl / dp;
				nd->min_height = min_height * dl / dp;
			}
			return 1;
		}
		return 0;

	case HID_UP_DIGITIZER:
		switch (usage->hid) {
		/* we do not want to map these for now */
		case HID_DG_CONTACTID: /* Not trustworthy, squelch for now */
		case HID_DG_INPUTMODE:
		case HID_DG_DEVICEINDEX:
		case HID_DG_CONTACTMAX:
			return -1;

		/* width/height mapped on TouchMajor/TouchMinor/Orientation */
		case HID_DG_WIDTH:
			hid_map_usage(hi, usage, bit, max,
				      EV_ABS, ABS_MT_TOUCH_MAJOR);
			input_set_abs_params(hi->input, ABS_MT_TOUCH_MAJOR,
					     field->logical_minimum,
					     field->logical_maximum,
					     /* 2mm fuzz */
					     20 * nd->sensor_logical_width /
					     nd->sensor_physical_width, 0);
			return 1;
		case HID_DG_HEIGHT:
			hid_map_usage(hi, usage, bit, max,
				      EV_ABS, ABS_MT_TOUCH_MINOR);
			input_set_abs_params(hi->input, ABS_MT_TOUCH_MINOR,
					     field->logical_minimum,
					     field->logical_maximum,
					     /* 2mm fuzz */
					     20 * nd->sensor_logical_height /
					     nd->sensor_physical_height, 0);
			input_set_abs_params(hi->input, ABS_MT_ORIENTATION,
					     0, 1, 0, 0);
			return 1;
		}
		return 0;

	case 0xff000000:
		/* we do not want to map these: no input-oriented meaning */
		return -1;
	}

	return 0;
}

static int ntrig_input_mapped(struct hid_device *hdev, struct hid_input *hi,
			      struct hid_field *field, struct hid_usage *usage,
			      unsigned long **bit, int *max)
{
	/* No special mappings needed for the pen and single touch */
	if (field->physical)
		return 0;

	if (usage->type == EV_KEY || usage->type == EV_REL
			|| usage->type == EV_ABS)
		clear_bit(usage->code, *bit);

	return 0;
}

static void ntrig_store_contact(struct hid_device *hdev, struct ntrig_data *nd)
{
	struct ntrig_contact *contact;
	struct ntrig_frame *frame;

	if (list_empty(&nd->old_contacts))
		hid_err(hdev, "Ran out of contacts\n");

	if (list_empty(&nd->old_contacts) || list_empty(&nd->frames))
		return;

	frame = list_first_entry(&nd->frames, struct ntrig_frame, list);

	contact = list_entry(nd->old_contacts.prev, struct ntrig_contact, list);
	contact->inactive = 0;
	contact->x = nd->x;
	contact->y = nd->y;
	contact->w = nd->w;
	contact->h = nd->h;
	contact->id = nd->maxcontacts;

	contact->dx = contact->dy = 0;

	contact->est_y_min = nd->y - nd->r_y;
	contact->est_y_max = nd->y + nd->r_y;
	contact->est_x_min = nd->x - nd->r_x;
	contact->est_x_max = nd->x + nd->r_x;

	list_move_tail(&contact->list, &frame->pending);
}

/* Update the state of a track to the most recent matched contact */
static inline void ntrig_update_track(struct ntrig_data *nd,
				      struct ntrig_contact *old,
				      struct ntrig_contact *cur,
				      struct ntrig_frame *frame)
{
	/* Simple motion estimation */
	cur->dx = cur->x - old->x;
	cur->dy = cur->y - old->y;
	cur->est_x_min = cur->x + cur->dx - nd->r_x;
	cur->est_x_max = cur->x + cur->dx + nd->r_x;
	cur->est_y_min = cur->y + cur->dy - nd->r_y;
	cur->est_y_max = cur->y + cur->dy + nd->r_y;

	/* Update the state of the contact in the frame */
	list_move_tail(&cur->list, &frame->tracked);

	/* Recycle the old contact.  At the moment only the
	 * most recent contact of a track is used. */
	old->inactive = nd->deactivate_slack;
	list_move_tail(&old->list, &nd->old_contacts);

	/* move the slot pointer to the new contact */
	cur->slot = old->slot;
	old->slot = NULL;

	/* Update the pointer in the active tracks list to the new
	 * contact.  If this track doesn't have a slot, assign it one
	 * and add to the tail of the list.  If we run out of slots
	 * we can assign one when a slot opens up. */
	if (cur->slot) {
		cur->id = cur->slot->id;
		list_replace(&old->active_tracks, &cur->active_tracks);
	} else if (!list_empty(&nd->available_slots)) {
		cur->slot = list_first_entry(&nd->available_slots,
				struct ntrig_slot, list);
		cur->id = cur->slot->id;
		list_del(&cur->slot->list);
		list_add_tail(&cur->active_tracks, &nd->active_tracks);
	}
}

static inline bool ntrig_match(struct ntrig_data *nd, struct ntrig_contact *a,
			       struct ntrig_contact *b,
			       struct ntrig_frame *frame)
{
	if (b->y >= a->est_y_min && b->y <= a->est_y_max &&
		b->x >= a->est_x_min && b->x <= a->est_x_max) {
		ntrig_update_track(nd, a, b, frame);
		return true;
	}

	return false;
}

static inline void ntrig_terminate_track(struct input_dev *input,
					 struct ntrig_data *nd,
					 struct ntrig_contact *contact)
{
	/* slot should not be NULL here, but checking to be sure */
	if (contact->slot) {
		/* termination events */
		input_mt_slot(input, contact->id);
		input_mt_report_slot_state(input, MT_TOOL_FINGER, false);

		/* return the slot to the available pool */
		list_add_tail(&contact->slot->list, &nd->available_slots);
		contact->slot = NULL;
		list_del(&contact->active_tracks);
	}

	/* mark this contact as expired */
	contact->inactive = nd->deactivate_slack;
}

/* Shift and increase the size of the region of the estimated position
 * for a lost track */
static inline bool ntrig_age_contact(struct input_dev *input,
				     struct ntrig_data *nd,
				     struct ntrig_contact *contact)
{
	contact->inactive++;
	//printk(KERN_ERR "Aging.  Current age: %d %d\n", contact->inactive, contact->slot!=NULL);
	if (contact->inactive >= nd->deactivate_slack) {
		contact->inactive++;
		ntrig_terminate_track(input, nd, contact);
		return true;
	} else {
		contact->est_x_min += contact->dx - (nd->r_x >> contact->inactive);
		contact->est_x_max += contact->dx + (nd->r_x >> contact->inactive);
		contact->est_y_min += contact->dy - (nd->r_y >> contact->inactive);
		contact->est_y_max += contact->dy + (nd->r_y >> contact->inactive);
	}
	return false;
}

#define WHERESTR  "[file %s, line %d]: "
#define WHEREARG  __FILE__, __LINE__
#define ntrig_loop_barf(l) if(l++ > 200) { printk(KERN_ERR "BARF %s: %d\n",__FILE__,__LINE__); break; }

static void track(struct input_dev *input, struct ntrig_data *nd) {
	struct ntrig_frame *prev_frame, *cur_frame;
	struct ntrig_contact *old, *old_tmp, *cur, *cur_tmp;

	if (list_empty(&nd->frames))
	{
		printk("no frames !!!\n");
		return;
	}

	prev_frame = list_first_entry(nd->frames.next, struct ntrig_frame, list);
	cur_frame = list_first_entry(&nd->frames, struct ntrig_frame, list);

	list_for_each_entry_safe(cur, cur_tmp, &cur_frame->pending, list) {
		/* First look for a match in current active tracks.
		 * By far the most common case is that this will
		 * match the first element in the list */
		list_for_each_entry_safe(old, old_tmp, &prev_frame->tracked, list) {
			if (ntrig_match(nd, old, cur, cur_frame))
				break;
		}

		/* Compare against retiring, but not yet deactivated tracks */
		list_for_each_entry_safe(old, old_tmp, &nd->old_contacts, list) {
			if (old->inactive >= nd->deactivate_slack &&
			    old->inactive)
				break;

			if (ntrig_match(nd, old, cur, cur_frame))
				break;
		}

		/* Search for a match among unmatch contacts in the previous
		 * frame.  This is where new tracks are found */
		list_for_each_entry_safe(old, old_tmp, &prev_frame->pending, list) {
			if (ntrig_match(nd, old, cur, cur_frame))
				break;
		}
	}

	/* Traces that were active last frame but absent from the current frame
	 * are dropped to the old_contacts list for aging and possible recovery */
	list_for_each_entry_safe(old, old_tmp, &prev_frame->tracked, list) {
		if (nd->deactivate_slack <= 0) {
			ntrig_terminate_track(input, nd, old);
			old->inactive = 1;
		}
	}
	list_splice_init(&prev_frame->tracked, &nd->old_contacts);
}

static void emit_frame(struct input_dev *input, struct ntrig_data *nd) {
	struct ntrig_contact *contact, *contact_tmp;
	struct ntrig_frame *frame;

	if (list_empty(&nd->frames))
		return;

	/* Age old contacts and terminate expired tracks */
	list_for_each_entry_safe(contact, contact_tmp, &nd->old_contacts, list) {
		/* Potentially interesting contacts should be
		 * sorted by the number of frames since they
		 * were last seen.  We can stop iterating with the
		 * first expired contact. */
		if (contact->inactive >= nd->deactivate_slack && contact->inactive)
			break;

		ntrig_age_contact(input, nd, contact);
	}

	frame = list_first_entry(&nd->frames,struct ntrig_frame, list);

	if (!list_empty(&frame->tracked)) {
		if (!list_empty(&nd->active_tracks)) {
			/* Emit the position of the oldest active track as
			 * normal events. */
			contact = list_first_entry(&nd->active_tracks,
						   struct ntrig_contact, active_tracks);
			input_report_key(input, BTN_TOUCH, 1);
			input_event(input, EV_ABS, ABS_X, contact->x -
				    (contact->dx / 2));
			input_event(input, EV_ABS, ABS_Y, contact->y -
				    (contact->dy / 2));
		}
	} else if (list_empty(&nd->active_tracks)) {
		input_report_key(input, BTN_TOUCH, 0);
	}

	/* Emit MT events */
	list_for_each_entry(contact, &frame->tracked, list) {
		if (!contact->slot)
			continue;

		input_mt_slot(input, contact->id);
		input_event(input, EV_ABS, ABS_MT_TRACKING_ID, contact->id);
		input_mt_report_slot_state(input, MT_TOOL_FINGER, true);
		input_event(input, EV_ABS, ABS_MT_POSITION_X, contact->x -
			    (contact->dx / 2));
		input_event(input, EV_ABS, ABS_MT_POSITION_Y, contact->y -
			    (contact->dy / 2));
		/*
		 * Translate from height and width to size
		 * and orientation.
		 */
		if (contact->w > contact->h) {
			input_event(input, EV_ABS, ABS_MT_ORIENTATION, 1);
			input_event(input, EV_ABS, ABS_MT_TOUCH_MAJOR,
				    contact->w);
			input_event(input, EV_ABS, ABS_MT_TOUCH_MINOR,
				    contact->h);
		} else {
			input_event(input, EV_ABS, ABS_MT_ORIENTATION, 0);
			input_event(input, EV_ABS, ABS_MT_TOUCH_MAJOR,
				    contact->h);
			input_event(input, EV_ABS, ABS_MT_TOUCH_MINOR,
				    contact->w);
		}
		//input_mt_sync(input);
	}


	list_rotate_right(&nd->frames);
	frame = list_first_entry(&nd->frames,struct ntrig_frame, list);

	/* Recycle the old conacts to prepare the frame for reuse */
	list_splice_init(&frame->tracked, &frame->pending);

	list_for_each_entry(contact, &frame->pending, list) {
		contact->inactive = nd->deactivate_slack;
	}
	list_splice_tail_init(&frame->pending, &nd->old_contacts);
}

/*
 * this function is called upon all reports
 * so that we can filter contact point information,
 * decide whether we are in multi or single touch mode
 * and call input_mt_sync after each point if necessary
 */
static int ntrig_event (struct hid_device *hid, struct hid_field *field,
			struct hid_usage *usage, __s32 value)
{
	struct ntrig_data *nd = hid_get_drvdata(hid);
	struct input_dev *input;

	/* Skip processing if not a claimed input */
	if (!(hid->claimed & HID_CLAIMED_INPUT))
		goto not_claimed_input;

	/* This function is being called before the structures are fully
	 * initialized */
	if (!(field->hidinput && field->hidinput->input))
		return -EINVAL;

	input = field->hidinput->input;

	/* No special handling needed for the pen */
	if (field->application == HID_DG_PEN)
		return 0;

	switch (usage->hid) {
	case 0xff000001:
		/* Tag indicating the start of a multitouch group */
		nd->reading_mt = 1;
		nd->first_contact_touch = 0;
		break;
	case HID_DG_TIPSWITCH:
		nd->tipswitch = value;
		/* Prevent emission of touch until validated */
		return 1;
	case HID_DG_CONFIDENCE:
		nd->confidence = value;
		break;
	case HID_GD_X:
		nd->x = value;
		/* Clear the contact footer */
		nd->mt_foot_count = 0;
		break;
	case HID_GD_Y:
		nd->y = value;
		break;
	case HID_DG_CONTACTID:
		nd->id = value;
		break;
	case HID_DG_WIDTH:
		nd->w = value;
		break;
	case HID_DG_HEIGHT:
		nd->h = value;
		/*
		 * when in single touch mode, this is the last
		 * report received in a finger event. We want
		 * to emit a normal (X, Y) position
		 */
		if (!nd->reading_mt) {
			/*
			 * TipSwitch indicates the presence of a
			 * finger in single touch mode.
			 */
			input_report_key(input, BTN_TOUCH,
					 nd->tipswitch);
			input_report_key(input, BTN_TOOL_DOUBLETAP,
					 nd->tipswitch);
			input_event(input, EV_ABS, ABS_X, nd->x);
			input_event(input, EV_ABS, ABS_Y, nd->y);
		}
		break;
	case 0xff000002:
		/*
		 * we receive this when the device is in multitouch
		 * mode. The first of the three values tagged with
		 * this usage tells if the contact point is real
		 * or a placeholder
		 */

		/* Shouldn't get more than 4 footer packets, so skip */
		if (nd->mt_foot_count >= 4)
			break;

		nd->mt_footer[nd->mt_foot_count++] = value;

		/* if the footer isn't complete break */
		if (nd->mt_foot_count != 4)
			break;

		/* Pen activity signal. */
		if (nd->mt_footer[2]) {
			/*
			 * When the pen deactivates touch, we see a
			 * bogus frame with ContactCount > 0.
			 * We can
			 * save a bit of work by ensuring act_state < 0
			 * even if deactivation slack is turned off.
			 */
			nd->act_state = deactivate_slack - 1;
			nd->confidence = 0;
			break;
		}

		/*
		 * The first footer value indicates the presence of a
		 * finger.
		 */
		if (nd->mt_footer[0]) {
			/*
			 * We do not want to process contacts under
			 * the size threshold, but do not want to
			 * ignore them for activation state
			 */
//			if (nd->w < nd->min_width ||
//			    nd->h < nd->min_height)
//				nd->confidence = 0;
		} else
			break;


		if (nd->confidence) {
			ntrig_store_contact(hid, nd);
		}
		break;

	case HID_DG_CONTACTCOUNT: /* End of a multitouch group */
		if (!nd->reading_mt) /* Just to be sure */
			break;

		nd->reading_mt = 0;

		track(input, nd);
		emit_frame(input, nd);

		break;

	default:
		/* fall-back to the generic hidinput handling */
		return 0;
	}

not_claimed_input:
	/* we have handled the hidinput part, now remains hiddev */
	if ((hid->claimed & HID_CLAIMED_HIDDEV) && hid->hiddev_hid_event)
		hid->hiddev_hid_event(hid, field, usage, value);

	return 1;
}

static int ntrig_alloc_mt_structures(struct input_dev *input, struct ntrig_data *nd)
{
	struct ntrig_frame *frames;
	struct ntrig_slot *slots;
	struct ntrig_contact *contacts;
	int i, num_frames, num_slots, num_contacts;

	num_frames = 3;
	num_slots = nd->maxcontacts * 2;
	num_contacts = (num_frames + 1) * nd->maxcontacts;

	frames = kcalloc(num_frames, sizeof(*frames), GFP_KERNEL);
	if (!frames)
		goto err;
	nd->first_frame = frames;

	contacts = kcalloc(num_contacts, sizeof(*contacts), GFP_KERNEL);
	if (!contacts)
		goto err_free_frames;
	nd->first_contact = contacts;

	slots = kcalloc(num_slots, sizeof(*slots), GFP_KERNEL);
	if (!slots)
		goto err_free_contacts;
	nd->first_slot = slots;

	if (input_mt_init_slots(input, num_slots))
		goto err_free_slots;

	input_set_abs_params(input, ABS_MT_SLOT, 0, nd->maxcontacts*2, 0, 0);
	input_set_abs_params(input, ABS_MT_TRACKING_ID, 0, nd->maxcontacts*2, 0, 0);

	/* Stuff the structures into their initial lists */
	for (i = 0; i < num_frames; i++) {
		INIT_LIST_HEAD(&frames[i].tracked);
		INIT_LIST_HEAD(&frames[i].pending);
		list_add(&frames[i].list, &nd->frames);
	}

	for (i = 0; i < num_slots; i++) {
		slots[i].id = i;
		list_add_tail(&slots[i].list, &nd->available_slots);
	}

	for (i = 0; i < num_contacts; i++) {
		/* Tag each as expired */
		contacts[i].inactive = nd->deactivate_slack;
		list_add(&contacts[i].list, &nd->old_contacts);
	}

	return 0;

err_free_slots:
	kfree(slots);
	nd->first_slot = NULL;
err_free_contacts:
	kfree(contacts);
	nd->first_contact = NULL;
err_free_frames:
	kfree(frames);
	nd->first_frame = NULL;
err:
	return -ENOMEM;
}

static int ntrig_probe(struct hid_device *hdev, const struct hid_device_id *id)
{
	int ret;
	struct ntrig_data *nd;
	struct hid_input *hidinput;
	struct input_dev *input;
	struct hid_report *report;

	if (id->driver_data)
		hdev->quirks |= HID_QUIRK_MULTI_INPUT
				| HID_QUIRK_NO_INIT_REPORTS;

	nd = kzalloc(sizeof(struct ntrig_data), GFP_KERNEL);
	if (!nd) {
		hid_err(hdev, "cannot allocate N-Trig data\n");
		return -ENOMEM;
	}

	nd->activate_slack = activate_slack;
	nd->act_state = activate_slack;
	nd->deactivate_slack = -deactivate_slack;
	nd->r_x = 500;
	nd->r_y = 400;

	/* Initialize tracking lists */
	INIT_LIST_HEAD(&nd->frames);
	INIT_LIST_HEAD(&nd->old_contacts);
	INIT_LIST_HEAD(&nd->available_slots);
	INIT_LIST_HEAD(&nd->active_tracks);

	hid_set_drvdata(hdev, nd);

	ret = hid_parse(hdev);
	if (ret) {
		hid_err(hdev, "parse failed\n");
		goto err_free;
	}

	ret = hid_hw_start(hdev, HID_CONNECT_DEFAULT & ~HID_CONNECT_FF);
	if (ret) {
		hid_err(hdev, "hw start failed\n");
		goto err_free;
	}


	/* This is needed for devices with more recent firmware versions */
	report = hdev->report_enum[HID_FEATURE_REPORT].report_id_hash[0x0a];
	if (report) {
		/* Let the device settle to ensure the wakeup message gets
		 * through */
		usbhid_wait_io(hdev);
		usbhid_submit_report(hdev, report, USB_DIR_IN);

		/*
		 * Sanity check: if the current mode is invalid reset it to
		 * something reasonable.
		 */
		if (ntrig_get_mode(hdev) >= 4)
			ntrig_set_mode(hdev, 3);
	}

	ntrig_report_version(hdev);

	/* If available poll for the MAXCONTACTS */
	report = hdev->report_enum[HID_FEATURE_REPORT].report_id_hash[4];
	if (report) {
		usbhid_submit_report(hdev, report, USB_DIR_IN);
		usbhid_wait_io(hdev);
		nd->maxcontacts = report->field[0]->value[0];
	}

	list_for_each_entry(hidinput, &hdev->inputs, list) {
		bool is_mt = false;
		int i;
		if (hidinput->report->maxfield < 1)
			continue;

		input = hidinput->input;
		report = hidinput->report;
		nd->id = 0;

		/*
		 * Determine if this is a mt input and count contactid fields.
		 * Any mt report will have at least 11 fields, 5 for each
		 * of 2 contacts and one for contact count.
		 */
		if (report->maxfield > 10) {
			for (i = 0; i < report->maxfield; i++)
				switch (report->field[i]->usage[0].hid) {
				case HID_DG_CONTACTCOUNT:
					is_mt = true;
					break;
				case HID_DG_CONTACTID:
					nd->id++;
					break;
				}
		}

		if (nd->id && !nd->maxcontacts)
			nd->maxcontacts = nd->id;

		switch (report->field[0]->application) {
		case HID_DG_PEN:
			input->name = "N-Trig Pen";
			break;
		case HID_DG_TOUCHSCREEN:
			/* These keys are redundant for fingers, clear them
			 * to prevent incorrect identification */
			__clear_bit(BTN_TOOL_PEN, input->keybit);
			__clear_bit(BTN_TOOL_FINGER, input->keybit);
			__clear_bit(BTN_0, input->keybit);
			__set_bit(BTN_TOOL_DOUBLETAP, input->keybit);
			if (!is_mt) {
				input->name = "N-Trig Touchscreen";
				break;
			}

			input->name = "N-Trig MultiTouch";
			ret = ntrig_alloc_mt_structures(input, nd);
			if (ret)
				goto err_free;
			break;
		}
	}

	ret = sysfs_create_group(&hdev->dev.kobj, &ntrig_attribute_group);

	return 0;
err_free:
	kfree(nd);
	return ret;
}

static void ntrig_remove(struct hid_device *hdev)
{
	struct ntrig_data *nd = hid_get_drvdata(hdev);
	struct hid_input *hidinput;

	sysfs_remove_group(&hdev->dev.kobj, &ntrig_attribute_group);
	hid_hw_stop(hdev);

	list_for_each_entry(hidinput, &hdev->inputs, list)
		if (hidinput->input->mtsize)
			input_mt_destroy_slots(hidinput->input);

	kfree(nd->first_frame);
	kfree(nd->first_slot);
	kfree(nd->first_contact);

	kfree(hid_get_drvdata(hdev));
}

static const struct hid_device_id ntrig_devices[] = {
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_1),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_2),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_3),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_4),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_5),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_6),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_7),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_8),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_9),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_10),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_11),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_12),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_13),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_14),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_15),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_16),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_17),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ HID_USB_DEVICE(USB_VENDOR_ID_NTRIG, USB_DEVICE_ID_NTRIG_TOUCH_SCREEN_18),
		.driver_data = NTRIG_DUPLICATE_USAGES },
	{ }
};
MODULE_DEVICE_TABLE(hid, ntrig_devices);

static const struct hid_usage_id ntrig_grabbed_usages[] = {
	{ HID_ANY_ID, HID_ANY_ID, HID_ANY_ID },
	{ HID_ANY_ID - 1, HID_ANY_ID - 1, HID_ANY_ID - 1 }
};

static struct hid_driver ntrig_driver = {
	.name = "ntrig",
	.id_table = ntrig_devices,
	.probe = ntrig_probe,
	.remove = ntrig_remove,
	.input_mapping = ntrig_input_mapping,
	.input_mapped = ntrig_input_mapped,
	.usage_table = ntrig_grabbed_usages,
	.event = ntrig_event,
};

static int __init ntrig_init(void)
{
	return hid_register_driver(&ntrig_driver);
}

static void __exit ntrig_exit(void)
{
	hid_unregister_driver(&ntrig_driver);
}

module_init(ntrig_init);
module_exit(ntrig_exit);
MODULE_LICENSE("GPL");
